﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.IO;

using OpenTK;
using OpenTK.Graphics.OpenGL;
using OpenTK.Input;


namespace CG1.Ex06.Rendering
{
    public enum ViewMode
    {
        OrthogonalScene,
        PerspectiveScene,
    }

    class World
    {
        #region Properties

        //Info: Camera settings
        Vector3 Origin = Vector3.Zero;
        Vector3 LookAt = Vector3.Zero;
        Vector3 Up = Vector3.UnitZ;
        public Double CameraFi = 45; 
        public Double CameraTheta = 45;
        public Double PerCamDis = 10;
        public Double OrtCamDis = 0.5;

        //Info: World settings
        public Int32 Width;
        public Int32 Height;

        
        #endregion

        #region Init the world

        public World(Int32 width, Int32 height)
        {
            Width = width;
            Height = height;
        }

        /// <summary>
        /// Will initialize OpenGL in World. Create Bufferes for Cube and Compile shaders
        /// </summary>
        public void InitGL()
        {
            //Initialize OpenGL
            GL.ShadeModel(ShadingModel.Smooth);
            GL.ClearColor(Color.FromArgb(57, 57, 57));
            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
            GL.Hint(HintTarget.PerspectiveCorrectionHint, HintMode.Nicest);
            GL.Enable(EnableCap.DepthTest);

            //Create viewport
            Viewport(ViewMode.PerspectiveScene);
        }

        /// <summary>
        /// Reshape will call Viewport() in current world and update all scenes
        /// </summary>
        public void ReshapeGL(Int32 width, Int32 height)
        {
            Width = width;
            Height = height;
            GL.Viewport(0, 0, Width, Height);
        }

        #endregion

        #region Render

        

        /// <summary>
        /// Render all scenes and copies of Cube
        /// </summary>
        public void Render()
        {
            //Clear color buffer and depth buffer
            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
            
            //Go to model matrix and set it to identity
            GL.MatrixMode(MatrixMode.Modelview);
            GL.LoadIdentity();
            //LightAndMaterial();
            UpdateCamera();

            GL.UseProgram(0);
            GL.Disable(EnableCap.Texture2D);
            RenderAxises();
        }

        /// <summary>
        /// Render basic Axises as lines. Very traditional type of rendering. Do not use in advanced projects.
        /// </summary>
        public void RenderAxises()
        {
            //ToDo: Render axises of our world.
        }

        #endregion

        #region Viewport and Camera

        /// <summary>
        /// Set view port place for rendering the content. Change projection due to current viewMode
        /// </summary>
        public void Viewport(ViewMode Mode)
        {
            GL.Viewport(0, 0, Width, Height);

            //Go to projection matrix
            GL.MatrixMode(MatrixMode.Projection);

            switch (Mode)
            {
                case ViewMode.OrthogonalScene:
                    SetOrtho();
                    break;
                case ViewMode.PerspectiveScene:
                    SetPerspective();
                    break;
                default:
                    GL.LoadIdentity();
                    break;
            }
        }

        private void SetOrtho()
        {
            OpenTK.Matrix4 projection = Matrix4.Identity;
            projection = OpenTK.Matrix4.CreateOrthographicOffCenter((float)OrtCamDis * -10, (float)OrtCamDis * 10, (float)OrtCamDis * -10, (float)OrtCamDis * 10, -1, 100);
            GL.LoadMatrix(ref projection);
        }

        private void SetPerspective()
        {
            OpenTK.Matrix4 projection;
            //Hint: Try to use different type of perspective. Or different 'fovy'.
            projection = OpenTK.Matrix4.CreatePerspectiveFieldOfView(MathHelper.PiOver2, Width / (float)Height, 0.1f, 200f);
            GL.LoadMatrix(ref projection);
        }

        /// <summary>
        /// Called when it is time to update camera Origin, LookAt and UpVector.
        /// it can be used for both orthogonal and perspective projections.
        /// </summary>
        public void UpdateCamera()
        {
            //Info: Clamp camera angles.
            if (CameraFi >= 360.0) CameraFi -= 360.0;
            if (CameraFi < 0) CameraFi += 360.0;
            if (CameraTheta >= 179.5) CameraTheta = 179.5;
            if (CameraTheta <= 0.5) CameraTheta = 0.5;

            //Info: Compute new Origin of camera.
            Origin.X = LookAt.X + (float)Math.Sin(Math.PI / 180 * CameraTheta) * (float)Math.Sin(Math.PI / 180 * CameraFi) * (float)PerCamDis;
            Origin.Y = LookAt.Y - (float)Math.Sin(Math.PI / 180 * CameraTheta) * (float)Math.Cos(Math.PI / 180 * CameraFi) * (float)PerCamDis;
            Origin.Z = LookAt.Z + (float)Math.Cos(Math.PI / 180 * CameraTheta) * (float)PerCamDis;

            //Info: Set new origin, looking direction and upper vector of camera
            Matrix4 modelview = Matrix4.LookAt(Origin, LookAt, Up);
            GL.LoadMatrix(ref modelview);
        }

        #endregion
    }
}
